"""Application."""
from __future__ import annotations

from dataclasses import dataclass
from dataclasses import replace
from pathlib import Path
from typing import Iterator
from typing import Optional

from . import locations
from .cache import Cache
from .compat import contextmanager
from .config import Config
from .engine import Engine
from .project import Project
from .template import Template


@dataclass
class Application:
    """Application."""

    config: Config
    cache: Cache

    @classmethod
    def create(
        cls, *, config_file: Optional[Path] = None, cache_dir: Optional[Path] = None
    ) -> Application:
        """Create an application."""
        if cache_dir is None:
            cache_dir = locations.cache

        if config_file is None:
            config_file = locations.config

        config = Config.load(config_file)
        cache = Cache(cache_dir)

        return cls(config=config, cache=cache)

    @contextmanager
    def load_template(
        self,
        location: str,
        *,
        directory: Optional[Path] = None,
        revision: Optional[str] = None
    ) -> Iterator[Template]:
        """Load a template."""
        location = self.config.abbreviations.expand(location)
        with self.cache.load(
            location, directory=directory, revision=revision
        ) as template:
            yield template

    @contextmanager
    def load_template_for_project(
        self,
        project: Project,
        *,
        directory: Optional[Path] = None,
        revision: Optional[str] = None
    ) -> Iterator[Template]:
        """Load a template."""
        with self.load_template(
            project.variables.location, directory=directory, revision=revision
        ) as template:
            variables = template.variables.override(project.variables)
            yield replace(template, variables=variables)

    def generate_project(
        self,
        template: Template,
        *,
        output_dir: Optional[Path] = None,
        interactive: bool = False,
        overwrite: bool = False
    ) -> Project:
        """Generate a project from a template."""
        if output_dir is None:
            output_dir = Path.cwd()

        engine = Engine(template, interactive=interactive, overwrite=overwrite)
        path = engine.generate(output_dir)

        return Project.create(path, name=template.name, version=template.version)

    def update_project(
        self,
        project: Project,
        *,
        interactive: bool = False,
        directory: Optional[Path] = None,
        revision: Optional[str] = None
    ) -> None:
        """Update a project with changes from its template."""
        with self.load_template_for_project(
            project, directory=directory, revision=revision
        ) as template:
            with project.update(
                name=template.name, version=template.version
            ) as worktree:
                engine = Engine(template, interactive=interactive, overwrite=True)
                engine.generate(worktree.parent)



"""Create a project."""
from pathlib import Path
from typing import Optional

from ..core.application import Application


def create(
    location: str,
    *,
    interactive: bool = True,
    revision: Optional[str] = None,
    directory: Optional[Path] = None,
    output_dir: Optional[Path] = None,
    config_file: Optional[Path] = None,
) -> None:
    """Create a project from a Cookiecutter template."""
    application = Application.create(config_file=config_file)

    with application.load_template(
        location, directory=directory, revision=revision
    ) as template:
        application.generate_project(
            template, output_dir=output_dir, interactive=interactive
        )



"""Command-line interface for updating projects."""
from typing import Optional

import click

from . import logging
from ..api import update as api
from ..core.utils import as_optional_path
from .errorhandler import errorhandler


@click.command()
@click.option(
    "--interactive / --no-interactive",
    default=False,
    show_default=True,
    help="Prompt for template variables.",
)
@click.option(
    "--revision",
    metavar="REV",
    help="Branch, tag, or commit hash of the template repository.",
)
@click.option(
    "--template-directory",
    metavar="DIR",
    help=(
        "Directory within the template repository that contains the "
        "cookiecutter.json file."
    ),
)
@click.option(
    "--config",
    metavar="FILE",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, allow_dash=True),
    envvar="CUTTY_CONFIG",
    help="Configuration file.",
)
def update(
    interactive: bool,
    revision: Optional[str],
    template_directory: Optional[str],
    config: Optional[str],
) -> None:
    """Update a project from a Cookiecutter template."""
    logging.configure()

    with errorhandler():
        api.update(
            interactive=interactive,
            revision=revision,
            directory=as_optional_path(template_directory),
            config_file=as_optional_path(config),
        )



"""Render."""
from pathlib import Path
from typing import Any
from typing import Dict

import jinja2

from . import extensions
from .template import Template


def create_environment(template: Template) -> jinja2.Environment:
    """Create the Jinja environment."""
    extensions_ = extensions.load(extra=template.variables.extensions)
    return jinja2.Environment(  # noqa: S701
        loader=jinja2.FileSystemLoader(str(template.repository)),
        extensions=extensions_,
        keep_trailing_newline=True,
        undefined=jinja2.StrictUndefined,
    )


class Renderer:
    """Renderer."""

    def __init__(self, template: Template) -> None:
        """Initialize."""
        self.template = template
        self.context: Dict[str, Any] = {}
        self.environment = create_environment(template)

    def bind(self, name: str, value: Any) -> None:
        """Assign a value to a template variable."""
        self.context[name] = value

    def render(self, value: str) -> str:
        """Render a string."""
        template = self.environment.from_string(value)
        return template.render(cookiecutter=self.context)

    def render_json(self, value: Any) -> Any:
        """Render a JSON value."""
        if isinstance(value, dict):
            return {
                self.render_json(key): self.render_json(val)
                for key, val in value.items()
            }

        if isinstance(value, list):
            return [self.render_json(item) for item in value]

        if isinstance(value, str):
            return self.render(value)

        if value is None:
            return None

        return str(value)

    def render_file(self, path: Path) -> str:
        """Render a file."""
        if path.is_absolute():
            path = path.relative_to(self.template.repository)

        template = self.environment.get_template(path.as_posix())
        return template.render(cookiecutter=self.context)



"""Project."""
from __future__ import annotations

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator

from . import exceptions
from . import git
from . import locations
from .compat import contextmanager
from .variables import Variables


logger = logging.getLogger(__name__)


@dataclass
class Project:
    """Project."""

    repository: git.Repository
    variables: Variables

    @classmethod
    def load(cls, path: Path) -> Project:
        """Load the project."""
        repository = git.Repository(path)
        variables = Variables.load(path / ".cookiecutter.json")

        return cls(repository, variables)

    @classmethod
    def create(cls, path: Path, *, name: str, version: str) -> Project:
        """Create the project."""
        repository = git.Repository.init(path)
        branch = f"{locations.name}/{name}"

        repository.add(all=True)
        repository.commit(message=f"Generate project from {name} {version}")
        repository.branch(branch)

        return cls.load(path)

    @contextmanager
    def update(self, *, name: str, version: str) -> Iterator[Path]:
        """Update the project."""
        branch = f"{locations.name}/{name}"
        self._ensure_branch_exists(branch)

        with self.repository.worktree(
            self.repository.path / ".git" / locations.name / self.repository.path.name,
            branch,
            checkout=False,
            force_remove=True,
        ) as worktree:
            yield worktree.path

            logger.debug(f"Worktree {worktree.path}")

            worktree.add(all=True)
            worktree.commit(message=f"Update {name} to {version}", verify=False)

        with exceptions.ProjectUpdateFailed():
            self.repository.cherrypick(branch, commit=False)
            self.repository.commit(edit=False)  # Run pre-commit hook

    def _ensure_branch_exists(self, branch: str) -> None:
        try:
            self.repository.rev_parse(branch, verify=True, quiet=True)
        except git.Error:
            (firstref,) = self.repository.rev_list("HEAD", max_count=1, max_parents=0)
            self.repository.branch(branch, firstref)



"""Engine."""
from pathlib import Path

from .generate import Generator
from .render import Renderer
from .survey import Survey
from .template import Template


class Engine:
    """Engine."""

    def __init__(
        self, template: Template, *, interactive: bool = True, overwrite: bool = False
    ) -> None:
        """Initialize."""
        self.renderer = Renderer(template)
        self.survey = Survey(template, renderer=self.renderer, interactive=interactive)
        self.generator = Generator(
            template, renderer=self.renderer, overwrite=overwrite
        )

    def generate(self, output_dir: Path) -> Path:
        """Generate the project."""
        self.survey.run()
        return self.generator.generate(output_dir)



"""Prompt."""
import json
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Optional

import click


def ask(name: str, *, default: Any) -> Any:
    """Prompt for name and return the entered value or given default."""
    return click.prompt(name, default=default)


def ask_choice(name: str, *, values: List[Any]) -> Any:
    """Prompt to choose from several values for the given name."""
    if not values:
        raise ValueError

    choices = {str(number): value for number, value in enumerate(values, 1)}
    numbers = choices.keys()
    prompt = "\n".join(
        [
            f"Select {name}:",
            *[f"{number} - {value}" for number, value in choices.items()],
            "Choose from {}".format(", ".join(numbers)),
        ]
    )
    choice = click.prompt(
        prompt, type=click.Choice(numbers), default="1", show_choices=False
    )

    return choices[choice]


def ask_json_dict(name: str, *, default: Dict[str, Any]) -> Dict[str, Any]:
    """Prompt to provide a dictionary of data."""

    def load_json_dict(value: Optional[str]) -> Any:
        """Load entered value as a JSON dict."""
        assert value is not None  # noqa: S101
        try:
            result = json.loads(value)
        except Exception:
            raise click.UsageError("Unable to decode to JSON.")

        if not isinstance(result, dict):
            raise click.UsageError("Requires JSON dict.")

        return result

    value = click.prompt(
        name, default="default", type=click.STRING, value_proc=load_json_dict
    )

    if value == "default":
        return default

    return cast(Dict[str, Any], value)



"""Hooks."""
from __future__ import annotations

import dataclasses
import subprocess  # noqa: S404
import sys
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator
from typing import Optional
from typing import TYPE_CHECKING

from . import exceptions
from .compat import contextmanager
from .utils import make_executable

if TYPE_CHECKING:
    from .render import Renderer


@contextmanager
def create_temporary_script(filename: str, text: str) -> Iterator[Path]:
    """Create script with given filename and contents, using a temporary directory."""
    with tempfile.TemporaryDirectory() as directory:
        script = Path(directory) / filename
        script.write_text(text)

        make_executable(script)

        yield script


def execute_script(script: Path, *, cwd: Path) -> None:
    """Execute a script from a working directory."""
    command = [Path(sys.executable), script] if script.suffix == ".py" else [script]
    shell = sys.platform == "win32"

    subprocess.run(command, shell=shell, cwd=cwd, check=True)  # noqa: S602


@dataclass
class Hook:
    """Hook."""

    path: Path

    def run(self, *, renderer: Renderer, cwd: Path) -> None:
        """Execute the hook from the specified directory."""
        with exceptions.ContentRenderError(self.path):
            text = renderer.render_file(self.path)

        with exceptions.HookFailed(self.path):
            with create_temporary_script(self.path.name, text) as script:
                execute_script(script, cwd=cwd)


@dataclass
class Hooks:
    """Hooks."""

    pre_gen_project: Optional[Hook]
    post_gen_project: Optional[Hook]

    @classmethod
    def load(cls, path: Path) -> Hooks:
        """Load the hooks."""
        return Hooks(*[cls.find(path, field.name) for field in dataclasses.fields(cls)])

    @classmethod
    def find(cls, hookdir: Path, name: str) -> Optional[Hook]:
        """Return the hook if found, or None."""
        if hookdir.is_dir():
            for path in hookdir.iterdir():
                if path.is_file() and path.stem == name and not path.name.endswith("~"):
                    return Hook(path)

        return None

    def run_pre_gen_project(self, *, renderer: Renderer, cwd: Path) -> None:
        """Run pre_gen_project hook."""
        if self.pre_gen_project is not None:
            self.pre_gen_project.run(renderer=renderer, cwd=cwd)

    def run_post_gen_project(self, *, renderer: Renderer, cwd: Path) -> None:
        """Run post_gen_project hook."""
        if self.post_gen_project is not None:
            self.post_gen_project.run(renderer=renderer, cwd=cwd)



"""Template."""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from . import exceptions
from .compat import cached_property
from .hooks import Hooks
from .variables import Variables


@dataclass(frozen=True)
class Template:
    """Template."""

    repository: Path
    name: str
    version: str
    variables: Variables
    hooks: Hooks

    @classmethod
    def load(cls, path: Path, *, name: str, version: str, location: str) -> Template:
        """Load the template variables."""
        variables = Variables.load(path / "cookiecutter.json", location=location)
        hooks = Hooks.load(path / "hooks")

        return cls(
            repository=path,
            name=name,
            version=version,
            variables=variables,
            hooks=hooks,
        )

    @cached_property
    def root(self) -> Path:
        """Return the template directory."""
        for path in self.repository.iterdir():
            if (
                path.is_dir()
                and "cookiecutter" in path.name
                and "{{" in path.name
                and "}}" in path.name
            ):
                return path

        raise exceptions.TemplateDirectoryNotFound(self.variables.location)



"""Abbreviations."""
from typing import Mapping

from . import exceptions


class Abbreviations:
    """Abbreviations for template locations."""

    def __init__(self) -> None:
        """Initialize."""
        self.abbreviations = {
            "gh": "https://github.com/{}.git",
            "gl": "https://gitlab.com/{}.git",
            "bb": "https://bitbucket.org/{}",
        }

    def update(self, abbreviations: Mapping[str, str]) -> None:
        """Update the registered abbreviations."""
        self.abbreviations.update(abbreviations)

    def expand(self, template: str) -> str:
        """Expand abbreviations in a template location."""
        if template in self.abbreviations:
            return self.abbreviations[template]

        prefix, _, rest = template.partition(":")
        abbreviation = self.abbreviations.get(prefix)
        if abbreviation is not None:
            with exceptions.InvalidAbbreviation(prefix, template, abbreviation):
                return abbreviation.format(rest)

        return template



"""Error handling."""
import sys
from typing import Iterator

from rich.console import Console

from ..core.compat import contextmanager
from ..core.exceptions import CuttyException


@contextmanager
def errorhandler() -> Iterator[None]:
    """Print errors and exit with a status of 1."""
    try:
        yield
    except CuttyException as error:
        sys.exit(str(error))
    except Exception:
        Console().print_exception()



"""Application cache."""
import hashlib
import logging
from pathlib import Path
from typing import Iterator
from typing import Optional

from . import exceptions
from . import git
from .compat import contextmanager
from .template import Template
from .versions import Version


logger = logging.getLogger(__name__)


DIGEST_SIZE = hashlib.blake2b.MAX_DIGEST_SIZE


def _hash(value: str) -> str:
    return hashlib.blake2b(value.encode(), digest_size=DIGEST_SIZE).hexdigest()


def _load_repository(location: str, path: Path) -> git.Repository:
    path /= git.Repository.name_from_location(location, bare=True)

    if not path.exists():
        with exceptions.CloneError(location):
            logger.debug(f"Clone {location} to {path}")

            return git.Repository.clone(location, path, mirror=True, quiet=True)

    with exceptions.UpdateError(location):
        logger.debug(f"Fetch {location} to {path}")

        repository = git.Repository(path)
        repository.update_remote(prune=True)
        return repository


class Cache:
    """Application cache."""

    def __init__(self, path: Path) -> None:
        """Initialize."""
        self.path = path

    @contextmanager
    def load(
        self,
        location: str,
        *,
        directory: Optional[Path] = None,
        revision: Optional[str] = None,
    ) -> Iterator[Template]:
        """Load a project template."""
        hash = _hash(location)
        path = self.path / "repositories" / hash[:2] / hash
        repository = _load_repository(location, path)
        version = Version.get(repository, revision=revision)
        worktree = path / "worktrees" / version.sha1

        logger.debug(f"Version {version.name} ({version.sha1})")

        with repository.worktree(
            worktree, version.sha1, detach=True, force_remove=True
        ):
            yield Template.load(
                worktree if directory is None else worktree / directory,
                location=location,
                name=repository.path.stem,
                version=version.name,
            )



"""Generate."""
import contextlib
import fnmatch
import logging
import os.path
import shutil
from pathlib import Path

from . import exceptions
from .render import Renderer
from .template import Template
from .utils import on_raise
from .utils import rmtree


logger = logging.getLogger(__name__)


class Generator:
    """Generator."""

    def __init__(
        self, template: Template, *, renderer: Renderer, overwrite: bool = False
    ) -> None:
        """Initialize."""
        self.template = template
        self.renderer = renderer
        self.overwrite = overwrite

    def generate(self, output_dir: Path) -> Path:
        """Generate project."""
        assert self.template.root.is_dir()  # noqa: S101

        with exceptions.PathRenderError(self.template.root):
            target = output_dir / self.renderer.render(self.template.root.name)

        if os.path.lexists(target):
            if not self.overwrite:
                raise exceptions.ProjectDirectoryExists(target)

            if not target.is_dir() or target.is_symlink():
                logger.debug(f"Remove {target}")
                target.unlink()

            cleanup = contextlib.nullcontext()
        else:
            cleanup = on_raise(rmtree, target)

        with cleanup:
            with exceptions.ProjectGenerationFailed():
                self._render_directory(self.template.root, target, root=True)

        return target

    def _render(self, source: Path, output_dir: Path) -> None:
        with exceptions.PathRenderError(source):
            name = self.renderer.render(source.name)

        if not name:
            return

        target = output_dir / name

        # If the target exists at this point, we are in overwrite mode. Remove
        # existing files and symlinks, but do not recursively remove directory
        # trees. Unlinking is required to replace symlinks, and to replace
        # regular files by symlinks or directories. For safety, a directory
        # tree is never replaced by a symlink or regular file. Furthermore,
        # generating a project _across_ a symlink is not supported.
        if target.exists() and not target.is_dir() or target.is_symlink():
            target.unlink()

        if source.is_symlink():
            self._render_symlink(source, target)
        elif source.is_dir():
            self._render_directory(source, target)
        else:
            self._render_file(source, target)

    def _render_directory(
        self, source: Path, target: Path, *, root: bool = False
    ) -> None:
        logger.debug(f"Generate {target}")

        target.mkdir(parents=True, exist_ok=True)
        shutil.copymode(source, target)

        if root:
            self.template.hooks.run_pre_gen_project(renderer=self.renderer, cwd=target)

        for entry in source.iterdir():
            self._render(entry, target)

        if root:
            self.template.hooks.run_post_gen_project(renderer=self.renderer, cwd=target)

    def _render_symlink(self, source: Path, target: Path) -> None:
        logger.debug(f"Generate {target}")

        source_target = os.readlink(source)
        with exceptions.SymlinkRenderError(source, source_target):
            target_target = self.renderer.render(source_target)

        target.symlink_to(target_target)
        shutil.copymode(source, target, follow_symlinks=False)

    def _render_file(self, source: Path, target: Path) -> None:
        logger.debug(f"Generate {target}")

        with exceptions.ContentRenderError(source):
            text = (
                self.renderer.render_file(source)
                if not self._is_copy_only(source)
                else source.read_text()
            )

        target.write_text(text)
        shutil.copymode(source, target)

    def _is_copy_only(self, source: Path) -> bool:
        source = source.relative_to(self.template.root)
        return any(
            fnmatch.fnmatch(str(path), pattern)
            for path in [source, *source.parents]
            for pattern in self.template.variables.copy_without_render
        )



"""Types."""
import contextlib
import os
import subprocess  # noqa: S404
from typing import Any
from typing import Mapping
from typing import TYPE_CHECKING
from typing import Union


if TYPE_CHECKING:  # pragma: no cover
    AbstractContextManager = contextlib.AbstractContextManager[Any]
    CompletedProcess = subprocess.CompletedProcess[str]
    PathLike = os.PathLike[str]
else:
    AbstractContextManager = contextlib.AbstractContextManager
    CompletedProcess = subprocess.CompletedProcess
    PathLike = os.PathLike

Context = Mapping[str, Any]
StrPath = Union[str, PathLike]



"""Locations."""
from pathlib import Path

import appdirs


name = "cutty"
cache = Path(appdirs.user_cache_dir(appname=name, appauthor=name))
config = Path(appdirs.user_config_dir(appname=name, appauthor=name)) / f"{name}.yml"



"""Git interface."""
from __future__ import annotations

import functools
import inspect
import itertools
import re
import shutil
import subprocess  # noqa: S404
import sys
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any
from typing import Callable
from typing import cast
from typing import Iterator
from typing import List
from typing import MutableMapping
from typing import Optional
from typing import TypeVar
from urllib.parse import urlparse

from . import exceptions
from .compat import cached_property
from .compat import contextmanager
from .types import CompletedProcess
from .types import StrPath
from .utils import removeprefix


class Error(Exception):
    """Git error."""

    @classmethod
    def from_subprocess(cls, error: subprocess.CalledProcessError) -> Error:
        """Create an error from subprocess.CalledProcessError."""
        command = " ".join(
            itertools.takewhile(lambda arg: not arg.startswith("-"), error.cmd[1:])
        )
        message = (
            error.stderr
            if error.stderr
            else f"returned non-zero exit status {error.returncode}"
        )

        return cls(command, message)

    def __init__(self, command: str, message: str) -> None:
        """Initialize."""
        self.command = command
        self.message = message

    def __str__(self) -> str:
        """Return a friendly error message, prefixed by the command."""
        return f"git {self.command}: {self.message}"


VERSION_PATTERN = re.compile(
    r"""
    (?P<major>\d+)\.
    (?P<minor>\d+)
    (\.(?P<patch>\d+))?
    """,
    re.VERBOSE,
)


@dataclass(frozen=True, order=True)
class Version:
    """Simplistic representation of git versions."""

    major: int
    minor: int
    patch: int
    _text: Optional[str] = field(default=None, compare=False)

    @classmethod
    def parse(cls, text: str) -> Version:
        """Extract major.minor[.patch] from the start of the text."""
        match = VERSION_PATTERN.match(text)

        if match is None:
            raise ValueError(f"invalid version {text!r}")

        parts = match.groupdict(default="0")

        return cls(
            int(parts["major"]), int(parts["minor"]), int(parts["patch"]), _text=text
        )

    def __str__(self) -> str:
        """Return the original representation."""
        return (
            self._text
            if self._text is not None
            else f"{self.major}.{self.minor}.{self.patch}"
        )


@dataclass
class Git:
    """Git program."""

    path: Path
    env: MutableMapping[str, str] = field(default_factory=dict)

    @classmethod
    def find(cls) -> Optional[Git]:
        """Find the Git program."""
        path = shutil.which("git")
        return None if path is None else cls(Path(path))

    @cached_property
    def version(self) -> Version:
        """Return the git version."""
        text = self.run("--version").stdout.strip()
        text = removeprefix(text, "git version ")
        return Version.parse(text)

    def check_version(self, minimum: str) -> bool:
        """Return True if Git has at least the given version."""
        version: Version = self.version  # mypy workaround for Python 3.7
        return Version.parse(minimum) <= version

    def run(self, *args: StrPath, cwd: Optional[Path] = None) -> CompletedProcess:
        """Invoke git."""
        try:
            return subprocess.run(  # noqa: S603
                [self.path, *args],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                env=self.env or None,
                cwd=cwd,
            )
        except subprocess.CalledProcessError as error:
            raise Error.from_subprocess(error) from error


class _GlobalContext:
    @cached_property
    def git(self) -> Git:
        """Return the git program."""
        git = Git.find()
        if git is None:
            raise exceptions.GitNotFound()
        return git


_global = _GlobalContext()


R = TypeVar("R")


def requires(version: str) -> Callable[[Callable[..., R]], Callable[..., R]]:
    """Document the minimum Git version."""

    def decorator(f: Callable[..., R]) -> Callable[..., R]:
        @functools.wraps(f)
        def wrapper(*args: Any, **kwargs: Any) -> R:
            if not _global.git.check_version(version):
                func_name = dict(inspect.getmembers(f)).get("__name__")
                raise exceptions.GitVersionRequired(
                    func_name, version, _global.git.version
                )
            return f(*args, **kwargs)

        return wrapper

    return decorator


T = TypeVar("T")


def _format_option(key: str, value: T) -> str:
    """Convert a key-value pair to a command-line option."""
    name = key.replace("_", "-")
    if isinstance(value, bool):
        return f"--{name}" if value else f"--no-{name}"
    else:
        return f"--{name}={value}"


def _format_options(**kwargs: Optional[T]) -> List[str]:
    """Convert keyword arguments to command-line options."""
    return [
        _format_option(key, value) for key, value in kwargs.items() if value is not None
    ]


class Repository:
    """Git repository."""

    def __init__(self, path: Optional[Path] = None) -> None:
        """Initialize."""
        self.path = path or Path.cwd()

    @classmethod
    @requires("1.5.6")
    def init(cls, path: Path) -> Repository:
        """Create a repository."""
        _global.git.run("init", cwd=path)

        return cls(path)

    @classmethod
    @requires("1.6.0")  # --mirror
    def clone(
        cls,
        location: str,
        path: Optional[Path] = None,
        *,
        quiet: Optional[bool] = None,
        mirror: Optional[bool] = None,
    ) -> Repository:
        """Clone a repository."""
        options = _format_options(quiet=quiet, mirror=mirror)

        if path is None:
            path = Path(cls.name_from_location(location, bare=mirror is True))

        _global.git.run("clone", *options, location, str(path))

        return cls(path)

    @classmethod
    def name_from_location(cls, location: str, *, bare: bool = False) -> str:
        """Return the directory name given the location."""
        url = urlparse(location)
        name = Path(url.path).stem
        return name if not bare else f"{name}.git"

    def git(self, *args: StrPath) -> CompletedProcess:
        """Invoke git."""
        process = _global.git.run(*args, cwd=self.path)
        if sys.version_info <= (3, 7):
            # Work around [no-any-return] with dataclasses on Python 3.7
            return cast(CompletedProcess, process)
        else:
            return process

    @requires("1.5.1")
    def update_remote(self, prune: Optional[bool] = None) -> None:
        """Fetch updates for remotes in the repository."""
        options = _format_options(prune=prune)
        self.git("remote", "update", *options)

    @requires("2.7.0")
    def get_remote_url(self, remote: str) -> str:
        """Retrieve the URL for a remote."""
        process = self.git("remote", "get-url", remote)
        return process.stdout.strip()

    @requires("1.6.1")  # --format ":short"
    def tags(self) -> List[str]:
        """Return the tags."""
        process = self.git("for-each-ref", "--format=%(refname:short)", "refs/tags")
        return process.stdout.split()

    @requires("2.9.0")  # --checkout
    def add_worktree(
        self,
        path: Path,
        ref: str,
        *,
        checkout: Optional[bool] = None,
        detach: Optional[bool] = None,
    ) -> Repository:
        """Add a worktree."""
        options = _format_options(checkout=checkout, detach=detach)
        self.git("worktree", "add", *options, str(path), ref)

        return Repository(path)

    @requires("2.17.0")
    def remove_worktree(self, path: Path, *, force: Optional[bool] = None) -> None:
        """Remove a worktree."""
        options = _format_options(force=force)
        self.git("worktree", "remove", *options, str(path))

    @contextmanager
    def worktree(
        self,
        path: Path,
        ref: str,
        *,
        checkout: Optional[bool] = None,
        detach: Optional[bool] = None,
        force_remove: Optional[bool] = None,
    ) -> Iterator[Repository]:
        """Context manager to add and remove a worktree."""
        worktree = self.add_worktree(path, ref, checkout=checkout, detach=detach)

        try:
            yield worktree
        finally:
            self.remove_worktree(path, force=force_remove)

    @requires("1.2.0")  # --short
    def rev_parse(
        self,
        rev: str,
        *,
        quiet: Optional[bool] = None,
        short: Optional[bool] = None,
        verify: Optional[bool] = None,
    ) -> str:
        """Return the SHA-1 hash for the given revision."""
        options = _format_options(quiet=quiet, short=short, verify=verify)
        process = self.git("rev-parse", *options, rev)
        return process.stdout.strip()

    @requires("1.7.5")  # --max-parents
    def rev_list(
        self,
        commit: str,
        *commits: str,
        max_count: Optional[int] = None,
        max_parents: Optional[int] = None,
    ) -> List[str]:
        """Lists commit objects in reverse chronological order."""
        options = _format_options(max_count=max_count, max_parents=max_parents)
        process = self.git("rev-list", *options, commit, *commits)
        return process.stdout.split()

    @requires("1.5.5")  # --exact-match
    def describe(
        self,
        ref: str,
        *,
        tags: Optional[bool] = None,
        exact_match: Optional[bool] = None,
    ) -> str:
        """Give an object a human readable name based on an available ref."""
        options = _format_options(tags=tags, exact_match=exact_match)
        process = self.git("describe", *options, ref)
        return process.stdout.strip()

    @requires("0.99.7")
    def add(self, *paths: StrPath, all: Optional[bool] = None) -> None:
        """Add file contents to the index."""
        options = _format_options(all=all)
        self.git("add", *options, "--", *paths)

    @requires("1.3.0")
    def rm(self, *paths: StrPath) -> None:
        """Remove files from the working tree and from the index."""
        self.git("rm", *paths)

    @requires("1.7.2")  # --allow-empty-message
    def commit(
        self,
        *paths: StrPath,
        allow_empty: Optional[bool] = None,
        allow_empty_message: Optional[bool] = None,
        edit: Optional[bool] = None,
        message: Optional[str] = None,
        verify: Optional[bool] = None,
    ) -> None:
        """Record changes to the repository."""
        options = _format_options(
            allow_empty=allow_empty,
            allow_empty_message=allow_empty_message,
            edit=edit,
            message=message,
            verify=verify,
        )
        self.git("commit", *options, "--", *paths)

    @requires("0.99.1")
    def branch(self, name: str, ref: Optional[str] = None) -> None:
        """Create a branch."""
        if ref is not None:
            self.git("branch", name, ref)
        else:
            self.git("branch", name)

    @requires("0.99")
    def tag(self, name: str, ref: Optional[str] = None) -> None:
        """Create a tag."""
        if ref is not None:
            self.git("tag", name, ref)
        else:
            self.git("tag", name)

    @requires("0.99.7")
    def merge(self, ref: str) -> None:
        """Join two development histories together."""
        self.git("merge", ref)

    @requires("0.99.6")
    def cherrypick(self, *commits: str, commit: Optional[bool] = None) -> None:
        """Apply the changes introduced by some existing commits."""
        options = _format_options(commit=commit)
        self.git("cherry-pick", *options, *commits)



"""Configuration."""
from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from pathlib import Path

import poyo

from . import exceptions
from . import locations
from .abbreviations import Abbreviations


@dataclass
class Config:
    """Configuration."""

    abbreviations: Abbreviations = field(default_factory=Abbreviations)

    @classmethod
    def load(cls, path: Path) -> Config:
        """Return the user configuration."""
        if path == locations.config and not path.exists():
            return cls()

        with exceptions.ConfigurationFileError(
            path
        ), exceptions.ConfigurationDoesNotExist(path).when(FileNotFoundError):
            text = path.read_text()

        with exceptions.InvalidConfiguration(path):
            data = poyo.parse_string(text)

        config = cls()

        if "abbreviations" in data:
            config.abbreviations.update(data["abbreviations"])

        return config



"""Update a project."""
from pathlib import Path
from typing import Optional

from ..core.application import Application
from ..core.project import Project


def update(
    *,
    interactive: bool = False,
    revision: Optional[str] = None,
    directory: Optional[Path] = None,
    config_file: Optional[Path] = None,
) -> None:
    """Update a project from a Cookiecutter template."""
    application = Application.create(config_file=config_file)
    project = Project.load(Path.cwd())

    application.update_project(
        project, directory=directory, revision=revision, interactive=interactive
    )



"""Main entry point of the command-line interface."""
import click

from .create import create
from .update import update


@click.group()
@click.version_option()
def main() -> None:
    """Cutty creates projects from templates."""


main.add_command(create)
main.add_command(update)



"""General-purpose utilities."""
import collections.abc
import contextlib
import functools
import importlib
import os
import shutil
import stat
from pathlib import Path
from typing import Any
from typing import Callable
from typing import ContextManager
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import TypeVar
from typing import Union

from .compat import contextmanager
from .types import AbstractContextManager


def as_optional_path(argument: Optional[str]) -> Optional[Path]:
    """Convert the argument to a Path if it is not None."""
    return Path(argument) if argument is not None else None


def removeprefix(string: str, prefix: str) -> str:
    """Remove prefix from string, if present."""
    return string[len(prefix) :] if string.startswith(prefix) else string


def rmtree(path: Path) -> None:
    """Remove a directory tree.

    On Windows, read-only files cannot be removed. Use the `onerror` callback
    to clear the read-only bit and retry.

    See https://docs.python.org/3/library/shutil.html#rmtree-example
    """

    def _onerror(function: Any, path: Any, excinfo: Any) -> Any:
        os.chmod(path, stat.S_IWRITE)
        function(path)

    shutil.rmtree(path, onerror=_onerror)


class OnRaise(AbstractContextManager):
    """Invoke a callback when leaving the context due to an exception."""

    def __init__(self, callback: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
        """Initialize."""
        self._args = (callback, args, kwargs)

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        """Exit the context."""
        if exc_type is not None:
            (callback, args, kwargs) = self._args
            callback(*args, **kwargs)


on_raise = OnRaise


def make_executable(path: Path) -> None:
    """Set the executable bit on a file."""
    status = os.stat(path)
    os.chmod(path, status.st_mode | stat.S_IEXEC)


@contextmanager
def multicontext(contexts: Iterable[ContextManager[Any]]) -> Iterator[List[Any]]:
    """Group multiple context managers in a single context manager."""
    with contextlib.ExitStack() as stack:
        yield [stack.enter_context(context) for context in contexts]


R = TypeVar("R")


def with_context(
    contextfactory: Callable[
        ..., Union[ContextManager[Any], Iterable[ContextManager[Any]]]
    ]
) -> Callable[[Callable[..., R]], Callable[..., R]]:
    """Invoke a function with a context manager created at call time."""

    def decorator(func: Callable[..., R]) -> Callable[..., R]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> R:
            context = contextfactory(*args, **kwargs)
            if isinstance(context, collections.abc.Iterable):
                context = multicontext(context)
            with context:
                return func(*args, **kwargs)

        return wrapper

    return decorator


def import_object(import_path: str) -> Any:
    """Import the object at the given import path.

    Import paths consist of the dotted module name, optionally followed by a
    colon or dot, and the module attribute at which the object is located.

    For example:

    - ``json``
    - ``os.path``
    - ``xml.sax.saxutils:escape``
    - ``xml.sax.saxutils.escape``

    This function mirrors the implementation of ``jinja2.utils.import_string``.
    """
    module_name, colon, attribute = import_path.rpartition(":")

    if not colon:
        module_name, _, attribute = import_path.rpartition(".")

    module = importlib.import_module(module_name)
    return getattr(module, attribute) if attribute else module



"""Survey."""
from typing import Any

from . import exceptions
from . import prompt
from .render import Renderer
from .template import Template
from .variables import Variable


class Survey:
    """Survey."""

    def __init__(
        self, template: Template, *, renderer: Renderer, interactive: bool = True
    ) -> None:
        """Initialize."""
        self.variables = sorted(
            template.variables, key=lambda variable: isinstance(variable.value, dict)
        )
        self.renderer = renderer
        self.interactive = interactive

    def run(self) -> None:
        """Bind variables from user input."""
        for variable in self.variables:
            value = self.ask(variable)
            self.renderer.bind(variable.name, value)

    def ask(self, variable: Variable) -> Any:
        """Return user input for a variable."""
        if variable.name.startswith("_"):
            return variable.value

        with exceptions.VariableRenderError(variable.name):
            value = self.renderer.render_json(variable.value)

        if not self.interactive:
            return value[0] if isinstance(variable.value, list) else value

        if isinstance(variable.value, list):
            return prompt.ask_choice(variable.name, values=value)

        if isinstance(variable.value, dict):
            return prompt.ask_json_dict(variable.name, default=value)

        return prompt.ask(variable.name, default=value)



"""Logging."""
import logging

from rich.logging import RichHandler


def configure() -> None:
    """Configure logging for console output."""
    logging.basicConfig(
        level="NOTSET", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
    )



"""Jinja2 extensions."""
import json
import string
from secrets import choice
from typing import Any
from typing import cast
from typing import Iterable
from typing import List
from typing import Optional
from typing import Type

import jinja2.ext
import jinja2_time
from slugify import slugify

from . import exceptions
from .utils import import_object


def jsonify(obj: Any) -> str:
    """Convert a Python object to JSON."""
    return json.dumps(obj, sort_keys=True, indent=4)


def random_ascii_string(length: int, punctuation: bool = False) -> str:
    """Create a random string."""
    corpus = string.ascii_letters
    if punctuation:
        corpus += string.punctuation
    return "".join(choice(corpus) for _ in range(length))


class JsonifyExtension(jinja2.ext.Extension):
    """Jinja2 extension to convert a Python object to JSON."""

    def __init__(self, environment: jinja2.Environment) -> None:
        """Initialize."""
        super().__init__(environment)
        environment.filters["jsonify"] = jsonify


class RandomStringExtension(jinja2.ext.Extension):
    """Jinja2 extension to create a random string."""

    def __init__(self, environment: jinja2.Environment) -> None:
        """Initialize."""
        super().__init__(environment)
        environment.globals["random_ascii_string"] = random_ascii_string


class SlugifyExtension(jinja2.ext.Extension):
    """Jinja2 extension to slugify a string."""

    def __init__(self, environment: jinja2.Environment) -> None:
        """Initialize."""
        super().__init__(environment)
        environment.filters["slugify"] = slugify


DEFAULT_EXTENSIONS = [
    JsonifyExtension,
    RandomStringExtension,
    SlugifyExtension,
    jinja2_time.TimeExtension,
]


def load_extension(import_path: str) -> Type[jinja2.ext.Extension]:
    """Import a Jinja extension from the specified path."""
    with exceptions.TemplateExtensionNotFound(import_path):
        extension = import_object(import_path)

    if not issubclass(extension, jinja2.ext.Extension):
        raise exceptions.TemplateExtensionTypeError(import_path, type(extension))

    return cast(Type[jinja2.ext.Extension], extension)


def load(*, extra: Optional[Iterable[str]] = None) -> List[Type[jinja2.ext.Extension]]:
    """Return the default Jinja extensions, plus any additional ones specified."""
    extensions = DEFAULT_EXTENSIONS[:]

    if extra is not None:
        extensions.extend(load_extension(path) for path in extra)

    return extensions



"""Determining the versions tagged in a Git repository."""
from __future__ import annotations

import contextlib
from dataclasses import dataclass
from typing import Iterator
from typing import Optional

import packaging.version

from . import exceptions
from . import git


@dataclass
class Tag:
    """Git tag for a version according to PEP 440."""

    name: str
    version: packaging.version.Version

    @classmethod
    def create(cls, name: str) -> Tag:
        """Parse the tag name as a PEP 440 version."""
        version = packaging.version.Version(name[1:] if name.startswith("v") else name)
        return cls(name, version)


def load_tags(repository: git.Repository) -> Iterator[Tag]:
    """Load versions tagged in the repository."""
    for tag in repository.tags():
        with contextlib.suppress(packaging.version.InvalidVersion):
            yield Tag.create(tag)


def find_latest_tag(repository: git.Repository) -> Optional[Tag]:
    """Return the Git tag for the latest version."""
    tags = load_tags(repository)

    with contextlib.suppress(ValueError):
        return max(tags, key=lambda tag: tag.version)

    return None


@dataclass
class Version:
    """Template version."""

    name: str
    sha1: str

    @classmethod
    def get(
        cls, repository: git.Repository, *, revision: Optional[str] = None
    ) -> Version:
        """Return the latest version, unless specified otherwise."""
        if revision is not None:
            return find_version(repository, revision)

        version = find_latest_version(repository)
        if version is not None:
            return version

        return find_version(repository, "HEAD")


def find_version(repository: git.Repository, revision: str) -> Version:
    """Return the tag or short SHA-1 for the revision."""
    with exceptions.InvalidRevision(revision):
        sha1 = repository.rev_parse(revision, verify=True)

    with contextlib.suppress(git.Error, packaging.version.InvalidVersion):
        tagname = repository.describe(revision, tags=True, exact_match=True)
        tag = Tag.create(tagname)
        return Version(tag.name, sha1)

    name = repository.rev_parse(revision, short=True)
    return Version(name, sha1)


def find_latest_version(repository: git.Repository) -> Optional[Version]:
    """Return the latest version tagged in the repository, if any."""
    tag = find_latest_tag(repository)

    if tag is None:
        return None

    sha1 = repository.rev_parse(tag.name, verify=True)
    return Version(tag.name, sha1)



"""Exceptions."""
from typing import Any
from typing import Iterator
from typing import Optional
from typing import Type

from .compat import contextmanager


class CuttyException(Exception):
    """Base class for cutty exceptions."""

    def __str__(self) -> str:
        """Convert the exception to a string."""
        assert self.__doc__ is not None  # noqa: S101
        message = self.__doc__.format(*self.args)
        if self.__cause__ is not None:
            message = message.rstrip(".")
            message = f"{message}: {self.__cause__}"
        return message

    def __enter__(self) -> None:
        """Enter the context."""

    def __exit__(
        self,
        exception_type: Optional[Type[BaseException]],
        exception: Optional[BaseException],
        traceback: Optional[Any],
    ) -> None:
        """Exit the context."""
        if isinstance(exception, Exception) and not isinstance(
            exception, CuttyException
        ):
            raise self from exception

    @contextmanager
    def when(self, *exception_types: Type[Exception]) -> Iterator[None]:
        """Context manager handling exceptions by raising self instead."""
        if not exception_types:
            exception_types = (Exception,)

        try:
            yield
        except CuttyException:
            raise
        except exception_types as exception:
            raise self from exception


class ConfigurationFileError(CuttyException):
    """Cannot read configuration file {}."""


class ConfigurationDoesNotExist(CuttyException):
    """Configuration file {} not found."""


class InvalidConfiguration(CuttyException):
    """The configuration file {} is not valid."""


class InvalidAbbreviation(CuttyException):
    """Invalid abbreviation {!r} in {!r}: {!r}."""


class GitNotFound(CuttyException):
    """Cannot find the git program."""


class GitVersionRequired(CuttyException):
    """{} requires git version {}, you have {}."""


class CloneError(CuttyException):
    """Cannot download template from {}."""


class UpdateError(CuttyException):
    """Cannot fetch template updates from {}."""


class InvalidRevision(CuttyException):
    """Invalid template revision {!r}."""


class TemplateDirectoryNotFound(CuttyException):
    """Template directory not found in {}."""


class TemplateConfigurationFileError(CuttyException):
    """Cannot read template configuration file {}."""


class TemplateConfigurationDoesNotExist(CuttyException):
    """Template configuration file {} not found."""


class InvalidTemplateConfiguration(CuttyException):
    """The template configuration file {} is not valid."""


class TemplateConfigurationTypeError(CuttyException):
    """The template configuration file {} is not valid: expected {}, but was {}."""


class MissingTemplateVariable(CuttyException):
    """Missing template variable {} in {}."""


class InvalidTemplateVariable(CuttyException):
    """Invalid template variable {} in {}: expected {}, but was {}."""


class TemplateExtensionNotFound(CuttyException):
    """Template extension {} not found."""


class TemplateExtensionTypeError(CuttyException):
    """Template extension {} is not an instance of jinja2.ext.Extension: {}."""


class RenderError(CuttyException):
    """Cannot render template."""


class VariableRenderError(RenderError):
    """Cannot render template variable {}."""


class PathRenderError(RenderError):
    """Cannot render template path {}."""


class ContentRenderError(RenderError):
    """Cannot render template {}."""


class SymlinkRenderError(RenderError):
    """Cannot render symlink {}: {}."""


class ProjectDirectoryExists(CuttyException):
    """Project directory {} already exists."""


class HookFailed(CuttyException):
    """Hook script {} failed."""


class ProjectGenerationFailed(CuttyException):
    """Project generation failed."""


class ProjectUpdateFailed(CuttyException):
    """Project update failed."""



"""Compatibility."""
import contextlib
import sys
from typing import Callable
from typing import ContextManager
from typing import Iterator
from typing import TYPE_CHECKING
from typing import TypeVar


T = TypeVar("T")


if sys.version_info >= (3, 8):
    from functools import cached_property as cached_property
elif TYPE_CHECKING:
    from typing import Any
    from typing import Type

    # fmt: off
    class cached_property:                                                # noqa: D101, N801
        def __init__(self, func: Callable[[Any], T]) -> None: ...         # noqa: D107, E704
        def __set_name__(self, owner: Type[Any], name: str) -> None: ...  # noqa: D105, E704
        def __get__(self, instance: Any, owner: Any = None) -> Any: ...   # noqa: D105, E704
    # fmt: on
else:
    from backports.cached_property import (  # noqa: F401
        cached_property as cached_property,
    )


def contextmanager(
    func: Callable[..., Iterator[T]]
) -> Callable[..., ContextManager[T]]:
    """Fix annotations of functions decorated by contextlib.contextmanager."""
    result = contextlib.contextmanager(func)
    result.__annotations__ = {**func.__annotations__, "return": ContextManager[T]}
    return result



"""Template variables."""
from __future__ import annotations

import contextlib
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional

from . import exceptions
from .utils import with_context


@dataclass(frozen=True)
class Variable:
    """Template variable."""

    name: str
    value: Any
    path: Path

    def as_string(self) -> str:
        """Check that the value is a string, and return it."""
        if not isinstance(self.value, str):
            raise exceptions.InvalidTemplateVariable(
                self.name, self.path.name, "str", repr(self.value)
            )

        return self.value

    def as_string_list(self) -> List[str]:
        """Check that the value is a list of strings, and return it."""
        if not (
            isinstance(self.value, list)
            and all(isinstance(item, str) for item in self.value)
        ):
            raise exceptions.InvalidTemplateVariable(
                self.name, self.path.name, "List[str]", repr(self.value)
            )

        return self.value


class Variables:
    """Collection of template variables."""

    @classmethod
    @with_context(
        lambda cls, path, **kwargs: (
            exceptions.TemplateConfigurationFileError(path.name),
            exceptions.TemplateConfigurationDoesNotExist(path.name).when(
                FileNotFoundError
            ),
            exceptions.InvalidTemplateConfiguration(path.name).when(
                json.decoder.JSONDecodeError
            ),
        )
    )
    def load(cls, path: Path, *, location: Optional[str] = None) -> Variables:
        """Load the template variables from a JSON file."""
        with path.open() as io:
            data = json.load(io)

        if not isinstance(data, dict):
            raise exceptions.TemplateConfigurationTypeError(
                path.name, "dict", type(data)
            )

        if location is not None:
            data["_template"] = location

        variables = [Variable(name, value, path) for name, value in data.items()]

        for variable in variables:
            if isinstance(variable.value, list) and not variable.value:
                raise exceptions.InvalidTemplateVariable(
                    variable.name, path.name, "non-empty list", repr(variable.value)
                )

        return cls(variables, path=path)

    def __init__(self, variables: Iterable[Variable], *, path: Path) -> None:
        """Initialize."""
        self.variables = {variable.name: variable for variable in variables}
        self.path = path

    def __iter__(self) -> Iterator[Variable]:
        """Iterate over the variables in the collection."""
        return iter(self.variables.values())

    def __getitem__(self, name: str) -> Variable:
        """Retrieve a variable by name."""
        return self.variables[name]

    def get(self, name: str, *, default: Any = None) -> Variable:
        """Return the variable named `name`, or the provided default."""
        with contextlib.suppress(KeyError):
            return self.variables[name]
        return Variable(name, default, self.path)

    @property
    def location(self) -> str:
        """Return the template location."""
        with exceptions.MissingTemplateVariable("_template", self.path.name):
            variable = self.variables["_template"]
        return variable.as_string()

    @property
    def extensions(self) -> List[str]:
        """Return the Jinja extensions."""
        variable = self.get("_extensions", default=[])
        return variable.as_string_list()

    @property
    def copy_without_render(self) -> List[str]:
        """Return patterns for files to be copied without rendering."""
        variable = self.get("_copy_without_render", default=[])
        return variable.as_string_list()

    def override(self, other: Variables) -> Variables:
        """Override variables from another collection."""

        def _override(variable: Variable) -> Variable:
            if variable.name.startswith("_"):
                return variable

            for other_variable in other:
                if variable.name == other_variable.name and (
                    not isinstance(variable.value, list)
                    or other_variable.value in variable.value
                ):
                    return other_variable

            # TODO: If variables are missing, we should prompt later.
            return variable

        return Variables([_override(variable) for variable in self], path=self.path)
